/*
 * WL24xx Wireless LAN Card Driver for FreeBSD
 *
 * References:
 *   1. WL24xx packet drivers (tooasm.asm)
 *   2. Access Point Firmware Interface Specification for IEEE 802.11 SUTRO
 *   3. IEEE 802.11
 *   4. Linux network driver (/usr/src/linux/drivers/net)
 *   5. WL24xx driver for Linux
 *   6. 3c505 FreeBSD driver (ISA:if_eg, PCMCIA:if_ed, if_ep)
 *
 * The structure of FreeBSD network drivers is more confused than Linux.
 * It doesn't split upper half from bottom layer. You will see lots of
 * threads mixed together.
 *
 * To catch the idea, I illustrate all program flows as follows:
 *
 * ISA bus:
 *
 *   Thread 1: zwl_probe_isa() -> zwl_attach() -> zwl_ioctl() -> zwl_init() -> zwl_start()
 *   Thread 2: zwlintr() --> zwl_rxintr() -> zwl_read() -> zwl_get()
 *                       \-> zwl_start()
 *   Thread 3: zwl_ioctl() -> zwl_setfilter()
 *   Thread 4: zwl_watchdog() -> zwl_reset()
 *
 * PCCARD (PCMCIA) bus:
 *
 *   Thread 1: zwl_init_pccard() -> zwl_probe_pccard() -> zwl_attach -> ...
 *   Thread 2: zwl_intr_pccard() -> zwlintr() -> ...
 *   Thread 3: zwl_unload_pccard() -> if_down()
 *   Reset refer ISA
 *
 * Tested with WL2420 firmware 1.2, FreeBSD 2.2.5-RELEASE
 */

/* #define ZWL_DEBUG */

/*
 * Should be determinated by "zwl.h" which is generated by 'config', 
 * but in order to distribute *.o only, I assign the Max value. Too Bad!!
 * Note: currently each zwl_softc is about 340 bytes.
 */
#include "zwl.h"
#undef  NZWL
#define NZWL    4

/* Always need pccard */
#include "card.h"
#undef  NCARD
#define NCARD   1

#define ZWL_NO_PROBE_ISA    0x0001
#define ZWL_DBG_IF          0x0010
#define ZWL_DBG_PC          0x0020
#define ZWL_DBG_PERF        0x0040
#define ZWL_DBG_ETH         0x0080

#ifdef  ZWL_DEBUG
#  define TRACE_IF(format, args...)  \
            { if (zwl_flags & ZWL_DBG_IF) printf(format , ## args); }
#  define TRACE_PC(format, args...)   \
            { if (zwl_flags & ZWL_DBG_PC) printf(format , ## args); }
#  define TRACE_PERF(format, args...)   \
            { if (zwl_flags & ZWL_DBG_PERF) printf(format , ## args); }
#  define TRACE_ETH(format, args...)   \
            { if (zwl_flags & ZWL_DBG_ETH) printf(format , ## args); }
#  define WHEN_ZWL_DEBUG(x) x
#else
#  define TRACE_IF(format, args...)
#  define TRACE_PC(format, args...)
#  define TRACE_PERF(format, args...)
#  define TRACE_ETH(format, args...)
#  define WHEN_ZWL_DEBUG(x)
#endif

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/kernel.h>         /* for ticks */

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef IPX
#include <netipx/ipx.h>
#include <netipx/ipx_if.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <machine/clock.h>

#include <i386/isa/isa_device.h>
#include <i386/isa/wlapi.h>
#include <i386/isa/wlpnp.h>

/*
 * Uncomment this line if you would like to provide wlu support.
 * Wlu (Unix version) use ioctl() to control WL24xx cards.
 */
#include <i386/isa/wlioctl.h>

#define ZWL_BUFLEN    0x0670

static int zwl_probe_isa(struct isa_device *);  /* for zwldriver */
static int zwl_attach(struct isa_device *);     /* for zwldriver */

/*
 * These are used at boot time by the configuration program.
 * Refer: /usr/src/sys/i386/isa/isa_device.h
 */
struct isa_driver zwldriver = {
    zwl_probe_isa,  /* test whether device is present */
    zwl_attach,     /* setup driver for a device */
    "zwl",          /* device name */
    0               /* true if other probes confuse us */
};

/*
 * Ethernet software status per interface.
 */
static struct zwl_softc {
    struct arpcom sc_arpcom;    /* Ethernet common part */
    WL_Adaptor  adaptor;

    u_char  *zwl_outbuf;        /* Outgoing packet buffer */
    u_char  gone;               /* PCCARD had been removed */

    WHEN_ZWL_DEBUG(DWORD   ulInterrupts); /* For debugging only */
    WHEN_ZWL_DEBUG(DWORD   ulRxNoMBuffer);
    WHEN_ZWL_DEBUG(DWORD   ulTxBcopyCalled);
    WHEN_ZWL_DEBUG(DWORD   ulRecvFunCalled);
    WHEN_ZWL_DEBUG(DWORD   ulZwlStartCalled);
    WHEN_ZWL_DEBUG(DWORD   ulZwlStartSendOk);
} zwl_softc[NZWL];

static const char *OP_MODE_TABLE[] = {
    "Access Point",
    "Infrastructure",
    "Ad-hoc",
    "Inter AP",
    NULL
};

static int zwl_flags;

/*
 * Definition for printing version information
 */
#ifdef WLAPI_DEBUG
#  define   WLAPI_DEBUG_STR         "WLAPI_DEBUG "
#else
#  define   WLAPI_DEBUG_STR         ""
#endif

#ifdef WL_SLOWDOWN_IO
#  define   WL_SLOWDOWN_IO_STR      "WL_SLOWDOWN_IO "
#else
#  define   WL_SLOWDOWN_IO_STR      ""
#endif

#ifdef ZWL_DEBUG
#  define   ZWL_DEBUG_STR           "ZWL_DEBUG "
#else
#  define   ZWL_DEBUG_STR           ""
#endif

#if NCARD > 0
#  define   NCARD_STR               "PCMCIA "
#else
#  define   NCARD_STR               ""
#endif

const char g_cpszVersion[] = "if_zwl.c: v1.20 " \
    WLAPI_DEBUG_STR WL_SLOWDOWN_IO_STR ZWL_DEBUG_STR NCARD_STR __DATE__;

int zwl_reset(struct zwl_softc *sc);            /* for wlioctl.c */

/*
 * BSD only provide bzero in kernel. :-(
 */
void* memset(void *pBuf, int nValue, size_t nSize)
{
    int i;
    for (i = 0; i < nSize; i++)
        ((BYTE *) pBuf)[i] = nValue;
        
    return pBuf;
}

/*
 * Convert mbufs to buf, avoid additional copy if possible.
 *
 * WARNING: the return buffer may not start with ptr
 */
static caddr_t zwl_mtobuf(struct mbuf *m0, caddr_t pExtraBuf) {
    struct  mbuf *m;
    caddr_t pBuf;

    /* A trick to prevent bcopy */
    if (m0->m_next == NULL) {
        /* only one m, send it directly */
        pBuf = mtod(m0, caddr_t);
    }
    else {
        caddr_t ptr = pExtraBuf;
        
        /* a mbuf link-list, copy to buffer first */
        for (m = m0; m != 0; m = m->m_next) {
            bcopy(mtod(m, caddr_t), ptr, m->m_len);
            ptr += m->m_len;
            WHEN_ZWL_DEBUG(sc->ulTxBcopyCalled++);
        }
            
        pBuf = pExtraBuf;
    }
    
    return pBuf;
}

/*
 * Start to remove mbuf from queue and deliver it to the SUTRO
 * Note: Because we cannot expect how many bytes available in card,
 *       we don't actually remove an mbuf from queue until the packet
 *       had been sent to card successfully.
 */

#define Splimp()    splimp()
#define Splx(s)     splx(s)

static void zwl_start(struct ifnet *ifp)
{
    register struct zwl_softc *sc = ifp->if_softc;
    struct  mbuf *m0;
    int     pktlen, s;
    caddr_t pBuf;

    WL_Adaptor *pAdaptor = &sc->adaptor;
    BOOL    bEnabled, bSend;

    WHEN_ZWL_DEBUG(sc->ulZwlStartCalled++);
    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
    
    s = Splimp();
    
    /* Don't transmit if interface is busy or not running */
    if ((ifp->if_flags & (IFF_RUNNING|IFF_OACTIVE)) != IFF_RUNNING) {
        Splx(s);
        return;
    }
    /* Set IFF_OACTIVE immediately after checking */
	ifp->if_flags |= IFF_OACTIVE;

    /* Sneak a peek at the next packet */
    m0 = ifp->if_snd.ifq_head;
    if (m0 == 0) {
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    /* We need to use m->m_pkthdr.len, so require the header */
    if ((m0->m_flags & M_PKTHDR) == 0)
        panic("zwl_start: no header mbuf");

    /* Try to padding the packet even if it's too short */
    pktlen = max(m0->m_pkthdr.len, ETHER_MIN_LEN);

    if (pktlen <= (int) sizeof(struct ether_header) || pktlen > ETHER_MAX_LEN) {
        TRACE_ETH("zwl#: Tx <<<<< Invliad Ethernet packet (size=%d)\n", pktlen);
        /* packet is obviously too large: toss it */
        IF_DEQUEUE(&ifp->if_snd, m0);
        m_freem(m0);
        ifp->if_oerrors++;
        ifp->if_flags &= ~IFF_OACTIVE;
        Splx(s);
        return;
    }

    TRACE_IF("zwl#: Tx .....\n");

    /* Mask interrupts from the SUTRO !! Do it as short as possible.        */
    /* We must mask interrupt from the same card, to prevent interrupt      */
    /* routine from accessing data structure and I/O port during WL_SendPkt */
    /* is running. It's very important!! */
    bEnabled = WL_BlockInterrupt(pAdaptor);

    pBuf = zwl_mtobuf(m0, sc->zwl_outbuf);

    /* Send the packet with default speed */
    bSend = WL_SendPkt(pAdaptor, pBuf, pktlen, 0);
    
    /* Turn SUTRO interrupt back on only if it is originally enabled */
    /* Must check first, because zwlintr() calls zwl_start() too! */
    if (bEnabled)
        WL_UnBlockInterrupt(pAdaptor);
    
    /* If sent successfully turn off IFF_OACTIVE. Otherwise, buffer is all used */
    /* IFF_OACTIVE will be 0 again when the SUTRO interrupts us and return ISR_Tx */
    if (bSend) {
        WHEN_ZWL_DEBUG(sc->ulZwlStartSendOk++);
        IF_DEQUEUE(&ifp->if_snd, m0);

#if NBPFILTER > 0
        if (ifp->if_bpf)
            bpf_mtap(ifp, m0);
#endif

        TRACE_IF("zwl#: Tx freem\n");
        m_freem(m0);
        
        ifp->if_timer = 2;  /* start timer, we must get ISR_Tx before timeout */
        ifp->if_flags &= ~IFF_OACTIVE;  /* Let others could own the transmitter */
    }
    /* I don't loop again. With field test, loop makes longer latency. */
    Splx(s);
}

/*
 * convert buf into mbufs
 */
static struct mbuf *zwl_get(struct zwl_softc *sc, int pktlen, int *pnRSSI)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    struct mbuf *top, **mprev, *m;
    int len, offset;

    ASSERT(pktlen >= 0);
    
    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m == NULL)
        return NULL;

    m->m_pkthdr.rcvif = ifp;
    m->m_pkthdr.len = pktlen;
    len = MHLEN;
    top = NULL;
    mprev = &top;
    offset = 0;

    while (pktlen > 0) {
        if (top) {
            MGET(m, M_DONTWAIT, MT_DATA);
            if (m == NULL) {
                TRACE_PERF("zwl#: No Rx mbuf available!\n");
                m_freem(top);
                return NULL;
            }
            len = MLEN;
        }
        if (pktlen >= MINCLSIZE) {
            MCLGET(m, M_DONTWAIT);
            if (m->m_flags & M_EXT)
                len = MCLBYTES;
        }
        m->m_len = len = min(pktlen, len);
        *pnRSSI = WL_ReceiveData(&sc->adaptor, offset, mtod(m, caddr_t), len);
        WHEN_ZWL_DEBUG(sc->ulRecvFunCalled++);
        offset += len;
        pktlen -= len;
        *mprev = m;
        mprev = &m->m_next;
    }

    ASSERT(top != NULL);
    return top;
}

/*
 * Pass a packet up to the higher levels.
 */
static int zwl_read(struct zwl_softc *sc, int pktlen)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    MADDR   *pAddr;
    
    struct mbuf *m;
    struct ether_header *eh;
    int bMorePkt;
    int nRSSI;
    
    if (pktlen <= (int) sizeof(struct ether_header) || pktlen > ETHER_MAX_LEN) {
        TRACE_ETH("zwl#: Rx <<<<< Invliad Ethernet packet (size=%d)\n", pktlen);
        ifp->if_ierrors++;
        
        /* skip this packet */
        return WL_ReceiveDataDone(pAdaptor);
    }
    
    /* Pull packet off interface. */
    m = zwl_get(sc, pktlen, &nRSSI);
    bMorePkt = WL_ReceiveDataDone(pAdaptor);
    
    if (m == NULL) {
        WHEN_ZWL_DEBUG(sc->ulRxNoMBuffer++);
        return bMorePkt;
    }

    ifp->if_ipackets++;

    /* We assume the header fit entirely in one mbuf. */
    eh = mtod(m, struct ether_header *);

    if (pAdaptor->mibExtra.CCR == CCR_AP) {
        MADDR *p = (MADDR *) eh->ether_dhost;
        /* If broadcast, multicast, or WL card, forward the packet */
        if ((p->b0 & 0x01) || (p->b0==0x00 && p->b1==0x60 && p->b2==0xB3 && 
            !MA_ISEQUAL(*p, pAdaptor->MacAddress))) {
            caddr_t pBuf = zwl_mtobuf(m, sc->zwl_outbuf);

            /* Drop the packet if send failed */
            if (!WL_SendPkt(pAdaptor, pBuf, pktlen, 0))
                ifp->if_oerrors++;
        }
    }

    /* Comes from monitored site ? */
    pAddr = (MADDR *) eh->ether_shost;
    if (MA_ISEQUAL(*pAddr, pAdaptor->Monitor.MAddr)) {
        pAdaptor->Monitor.RSSI = nRSSI;
        pAdaptor->Monitor.TimeStamp = ticks;
    }
    
#if NBPFILTER > 0
    /*
     * Check if there's a BPF listener on this interface.
     * If so, hand off the raw packet to BPF.
     */
    if (ifp->if_bpf) {
        bpf_mtap(ifp, m);

        /*
         * Note that the interface cannot be in promiscuous mode if
         * there are no BPF listeners.  And if we are in promiscuous
         * mode, we have to check if this packet is really ours.
         */
        if ((ifp->if_flags & IFF_PROMISC) &&
            (eh->ether_dhost[0] & 1) == 0 && /* !mcast and !bcast */
            bcmp(eh->ether_dhost, sc->sc_arpcom.ac_enaddr,
                sizeof(eh->ether_dhost)) != 0) {
            m_freem(m);
            return bMorePkt;
        }
    }
#endif

    /* We assume the header fit entirely in one mbuf. */
    m_adj(m, sizeof(struct ether_header));
    ether_input(ifp, eh, m);
    
    return bMorePkt;
}

static void zwl_rxintr(struct zwl_softc *sc, BYTE uStatus2)
{
    int         nSize;
    BOOL        bMorePkts;
    WL_Adaptor *pAdaptor = &sc->adaptor;
    
    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    if (uStatus2 & ISR2_RxMgmt) {
        TRACE_IF("zwl#: Rx >>>>> Roaming\n");
        WL_Roaming(pAdaptor);
    }
    
    /* Check for data queue */
    bMorePkts = TRUE;
    while (bMorePkts && (nSize = WL_ReceiveDataLook(pAdaptor)) != -1) {
        TRACE_IF("zwl#: Rx >>>>> Data\n");
        /* Pass a packet up to the higher levels */
        bMorePkts = zwl_read(sc, nSize);
    }
}

void zwlintr(int unit)
{
    struct zwl_softc *sc = &zwl_softc[unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    
    WL_Adaptor *pAdaptor = &sc->adaptor;
    CSB_BLOCK   csb;
    BYTE        cTmp;
    BYTE        cStatus;
    int         x;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
    
    x = splhigh();
    
    /* Ack Interrupt */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);

    WL_BlockInterrupt(pAdaptor);
    
    /* For debugging */
    TRACE_IF("WL: interrupt %lu begin...\n", ++sc->ulInterrupts);
            
    /* Clean Zone Begin */
    WL_LockSutro(pAdaptor);
    
    WL_GetFromWla(pAdaptor, Control_Status_Block, &csb, sizeof(csb));
    
    cTmp = 0;
    WL_SetToWla(pAdaptor, CSB_Interrupt_Status, &cTmp, sizeof(cTmp));
    WL_SetToWla(pAdaptor, CSB_Interrupt_Status2, &cTmp, sizeof(cTmp));

    WL_UnlockSutro(pAdaptor);

    /* 0x80 = Stop driver, and 0x40 = Restart driver are reserved */
    cStatus = csb.Interrupt_Status & 0x3F;

    /* Check ISR_Tx first, to increase network throughput */    
    if (cStatus & ISR_Tx) {
        TRACE_IF("zwl#: Tx <<<<<\n");
        WL_PurgeTxQueue(pAdaptor, &pAdaptor->qData);
        ifp->if_opackets++;
        ifp->if_timer = 0;  /* reset timer */
        ifp->if_flags &= ~IFF_OACTIVE;

        /* Verify zwl_start() carefully */
        zwl_start(ifp);
    }
    
    if (cStatus & ISR_Rx) {
        zwl_rxintr(sc, csb.Interrupt_Status2);
    }
    
    /* Clean Zone End */
    WL_UnBlockInterrupt(pAdaptor);

    splx(x);    
    return;
}

static void zwl_init(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    BYTE        cFilter, cCCR;
    int         s;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;

    /* For PCCARD */        
    s = splhigh();

    /* Initial WL24 firmware */
    printf("zwl%d: Initialize WL24xx firmware...\n", ifp->if_unit);
    if (!WL_InitFirmware(pAdaptor)) {
        printf("zwl%d: Cannot Initial Firmware!!\n", ifp->if_unit);
        /* Free IRQ, and mark IRQ as unused */
        return;
    }

    /* Print card setup information */
    cCCR = pAdaptor->mibExtra.CCR;

    printf("zwl%d: Channel %d, Speed %d Mb/s, %s Mode", ifp->if_unit,
        (int) pAdaptor->mibExtra.Channel, (int) pAdaptor->mibExtra.Speed,
        cCCR <= CCR_IAP ? OP_MODE_TABLE[cCCR] : "UNKNOWN");

    if (cCCR != CCR_ADHOC) {    /* should be better within one printf :-) */
        printf(", ESS ID = %c%c%c%c%c%c", 
            pAdaptor->mibExtra.ESS_ID.b0, pAdaptor->mibExtra.ESS_ID.b1,
            pAdaptor->mibExtra.ESS_ID.b2, pAdaptor->mibExtra.ESS_ID.b3,
            pAdaptor->mibExtra.ESS_ID.b4, pAdaptor->mibExtra.ESS_ID.b5);
    }
    printf("\n");

    /* Set up receiver filter to accept UNICAST and BROADCAST packets */
    cFilter = RMR_UNICAST | RMR_BROADCAST;
    WL_SetMibValue(pAdaptor, TYPE_EXTRA_MIB, IDX_RECEIVEMODE, &cFilter, sizeof(cFilter));

    if (sc->zwl_outbuf == NULL)
        sc->zwl_outbuf = malloc(ZWL_BUFLEN, M_TEMP, M_NOWAIT);

    /* Interface is now `running', with no output active. */
    ifp->if_flags |= IFF_RUNNING;
    ifp->if_flags &= ~IFF_OACTIVE;

    /* Acknowledge Interrupt, for cleaning last state */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* Enable interrupt from card after all */
    WL_UnBlockInterrupt(pAdaptor);
    
    printf("zwl%d: device opened\n", ifp->if_unit);

    /* Attempt to start output, if any. */
    zwl_start(ifp);
    
    splx(s);
}

static void zwl_stop(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    /* Ack Interrupt */
    OutB(GCR_ECINT, pAdaptor->nBaseAddr + NIC_GCR);

    /* Mask interrupts from the SUTRO */
    WL_BlockInterrupt(pAdaptor);
    
    /* Free IRQ, and mark IRQ as unused */
    
    printf("zwl%d: device closed\n", ifp->if_unit);
}

/*
 * Reset the SUTRO. Just simply stop and initial again
 */ 
int zwl_reset(struct zwl_softc *sc)
{
    int s;
    
    if (sc->gone)   /* For PCCARD: it had been removed */
        return -EIO;
        
    s = splimp();
    
    zwl_stop(sc);
    zwl_init(sc);
    
    splx(s);
    return 0;
}

static void zwl_watchdog(struct ifnet *ifp)
{
    struct zwl_softc *sc = ifp->if_softc;

    if (sc->gone)   /* For PCCARD: it had been removed */
        return;
        
    log(LOG_ERR, "zwl%d: device timeout\n", ifp->if_unit);
    sc->sc_arpcom.ac_if.if_oerrors++;

    zwl_reset(sc);
}

static void zwl_setfilter(struct zwl_softc *sc)
{
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    BYTE cFilter;
    int s;

    if (ifp->if_flags & IFF_PROMISC) {
        /* Promiscuous mode */
        cFilter = RMR_UNICAST | RMR_PROMISCUOUS | RMR_ALL_MULTICAST | RMR_BROADCAST;
    }
    else if (sc->sc_arpcom.ac_multicnt > 0) {
        /* Allow multicast */
        cFilter = RMR_UNICAST | RMR_ALL_MULTICAST | RMR_BROADCAST;
    }
    else {
        /* Normal mode */
        cFilter = RMR_UNICAST | RMR_BROADCAST;
    }
    
    /* Mask all SW/HW interrupts) */
    s = splhigh();

    /* Must not be interrupted */
    WL_SetMibValue(pAdaptor, TYPE_EXTRA_MIB, IDX_RECEIVEMODE, &cFilter, sizeof(cFilter));
    
    splx(s);
}

static int zwl_ioctl(register struct ifnet *ifp, int command, caddr_t data)
{
    struct zwl_softc *sc = ifp->if_softc;
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int s, error = 0;

    TRACE_PERF("zwl#: ulInterrupts=%10lu,\tulRxNoMBuffer=%lu\n"
               "zwl#: ulTxBcopyCalled=%10lu,\tulRecvFunCalled=%lu\n"
               "zwl#: ulZwlStartCalled=%10lu,\tulZwlStartSendOk=%lu",
                sc->ulInterrupts, sc->ulRxNoMBuffer, 
                sc->ulTxBcopyCalled, sc->ulRecvFunCalled,
                sc->ulZwlStartCalled, sc->ulZwlStartSendOk);

    if (sc->gone) { /* For PCCARD: it had been removed */
        ifp->if_flags &= ~IFF_RUNNING;
        return ENXIO;
    }
    
    s = splnet();

    switch (command) {

    case SIOCSIFADDR:
    case SIOCGIFADDR:
        ether_ioctl(ifp, command, data);
        break;

    case SIOCSIFFLAGS:
        if ((ifp->if_flags & IFF_UP) == 0 &&
            (ifp->if_flags & IFF_RUNNING) != 0) {
            /* If interface is marked down and it is running, then stop it. */
            zwl_stop(sc);
            ifp->if_flags &= ~IFF_RUNNING;
        } else if ((ifp->if_flags & IFF_UP) != 0 &&
               (ifp->if_flags & IFF_RUNNING) == 0) {
            /* If interface is marked up and it is stopped, then start it. */
            TRACE("zwl#: call zwl_init in zwlioctl\n");
            zwl_init(sc);
        } else {
            /* Deal with flags changes: IFF_MULTICAST, IFF_PROMISC */
            zwl_setfilter(sc);
        }
        break;

#ifdef __WL_IOCTL_H__
    case SIOC_WL_IOCTL: {
        /* Must block all interrupts */
        int hi = splhigh();
        struct ifreq *ifr = (struct ifreq *)data;
        error = -WL_IoctlWla(pAdaptor, (WL_IOCTL_BLK *) &ifr->ifr_data, sc);
        splx(hi);
        break;
    }
#endif

    default:
        error = EINVAL;
        break;
    }

    splx(s);
    return error;
}

static int zwl_attach(struct isa_device *isa_dev)
{
    int unit = isa_dev->id_unit;
    struct zwl_softc *sc = &zwl_softc[unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;
    int attached = (ifp->if_softc != NULL);
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int i;

    /* print probe information */
    printf("zwl%d: WL24xx at 0x%3.3x, MAC address in flash ROM:", 
            unit, pAdaptor->nBaseAddr);
    for (i = 0; i < 6; i++) {
        printf("%c%02x", i ? ':' : ' ', ((BYTE *) &pAdaptor->MacAddress)[i]);
    }
    printf("\n");

    bcopy(&pAdaptor->MacAddress, sc->sc_arpcom.ac_enaddr, ETHER_ADDR_LEN);
	
    /* Initialize ifnet structure. */
    ifp->if_softc = sc;
    ifp->if_unit = unit;
    ifp->if_name = "zwl";
    ifp->if_output = ether_output;
    ifp->if_start = zwl_start;
    ifp->if_ioctl = zwl_ioctl;
    ifp->if_watchdog = zwl_watchdog;
    ifp->if_init = (if_init_f_t*) zwl_init;
    ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX;
    
    /* Must have this, otherwise you'll get ENOBUFS in sendto(2) */
    /* It is specially for PCCARD, that doesn't initialize it.   */
    /* Refer: ifinit() in net/if.c (ifqmaxlen declared there)    */
    /* After 1 day's debugging... :-(                            */
    if (ifp->if_snd.ifq_maxlen == 0 )
        ifp->if_snd.ifq_maxlen = ifqmaxlen;
    
    /* Now we can attach the interface. (For PCCARD, must test first) */
    /* XXX: After 2 days debuging...                                  */
    if (!attached) {
        if_attach(ifp);
        ether_ifattach(ifp);
#if NBPFILTER > 0
        bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif
    }

    TRACE("zwl%d: zwl_attach ok!\n", unit);
    return 1;
}

/*
 * Probe if there is any ISA WL24xx card.
 * Return: 0 if failed, or occupied I/O region
 *
 * To understand ISA probing procdure, refer /usr/src/sys/i386/isa/isa.c
 * CAUTION: id_irq is bitmask (refer /usr/src/sys/i386/isa/icu.h)
 */
static int zwl_probe_isa(struct isa_device * id)
{
    struct zwl_softc *sc = &zwl_softc[id->id_unit];
    int  nCardIndex = id->id_unit + 1;      /* start with 1 */
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int         i;

    /* Check if byte alignment is correct or not. FOR DEBUGGING ONLY */
    if (sizeof(TX_HEADER) != 54 || sizeof(BEACON_PKT) != 55) {
        printf("zwl_probe_isa.c: compiler alignment option must be 1 byte.\n");
        return 0;
    }

    /* device zwl0 at isa? flags 0x1 net vector zwlintr */
    zwl_flags = id->id_flags;
    
    /* higher 16 bits as wlapi debug flags */
    WHEN_WLAPI_DEBUG(WL_DbFlag = zwl_flags >> 16);   
    
    if (id->id_flags & ZWL_NO_PROBE_ISA) {
        /* Specically for PCCARD */
        printf("zwl%d: Skip WL24xx ISA Probe (flags=0x%x)\n", id->id_unit, id->id_flags);
        return 0;
    }
    
    /* port ?       => id_iobase = 0xffffffff = -1 */
    /* port 0xffff  => id_iobase = 0xffff */
    /* not assigned => id_iobase = 0 */
    if (id->id_iobase <= 0 || id->id_iobase == 0xffff || id->id_irq == 0) {
        DWORD   ulBaseAddr;
        BYTE    cIRQ;

        printf("zwl%d: Trying to probe WL24xx PnP card...\n", id->id_unit);
        if (WL_PnpProbe(nCardIndex, &ulBaseAddr, &cIRQ)) {
            id->id_iobase   = ulBaseAddr;
            id->id_irq      = 1 << cIRQ;
            printf("zwl%d: WL24xx PnP card probe at 0x%3.3x, IRQ %d\n", 
                    id->id_unit, (int) id->id_iobase, (int) cIRQ);
        }
        else {
            printf("zwl%d: Cannot probe PnP card, you must specify base address and irq.\n", id->id_unit);
            return 0;
        }
    }
    else {
        /* Try to find non-PNPed WL24xx PnP cards that are installed */
        /* in a PC which does not support PnP BIOS. */
        /* If found, set the card's IRQ and I/O up. e.g. ffs(0x0010)=5 */
        WL_PnpSetNonInitCard(id->id_iobase, ffs(id->id_irq)-1);
    }    
    
    sc->gone = 0;
    bzero(pAdaptor, sizeof(*pAdaptor));
    pAdaptor->nBaseAddr = id->id_iobase;

    if (!WL_GetFlashMacAddress(pAdaptor)) {
        printf("zwl%d: Cannot read MAC address in flash ROM ??\n", id->id_unit);
        return 0;
    }

    printf("zwl%d: ISA card probed\n", id->id_unit);
    return WL_IO_REGION;
}


#if NCARD > 0
/*
 *  PC-Card (PCMCIA) specific code.
 */
#include <sys/select.h>
#include <pccard/card.h>
#include <pccard/driver.h>
#include <pccard/slot.h>
#include <sys/kernel.h>     /* for DATA_SET() */

static int  zwl_init_pccard(struct pccard_devinfo *);   /* Initial device */
static void zwl_unload_pccard(struct pccard_devinfo *); /* Disable driver */
static int  zwl_intr_pccard(struct pccard_devinfo *);   /* Interrupt handler */

static struct pccard_device zwl_info = {
    "zwl",
    zwl_init_pccard,
    zwl_unload_pccard,
    zwl_intr_pccard,
    0,                  /* Attributes - presently unused */
    &net_imask          /* Interrupt mask for device */
                        /* XXX - Should this also include net_imask? */
};

DATA_SET(pccarddrv_set, zwl_info);

/*
 * Probe if there is any PCCARD WL24xx card.
 * Return: 0 if failed, or occupied I/O region
 *
 * Do what zwl_probe() does, but skip PnP functions
 */
static int zwl_probe_pccard(struct isa_device * id)
{
    struct zwl_softc *sc = &zwl_softc[id->id_unit];
    int  nCardIndex = id->id_unit + 1;      /* start with 1 */
    
    WL_Adaptor  *pAdaptor = &sc->adaptor;
    int         s;

    sc->gone = 0;
    bzero(pAdaptor, sizeof(*pAdaptor));
    pAdaptor->nBaseAddr = id->id_iobase;

    TRACE_PC("zwl#: WL_GetFlashMacAddress()\n");
    s = splbio();
    if (!WL_GetFlashMacAddress(pAdaptor)) {
        printf("zwl%d: Cannot read MAC address in flash ROM ??\n", id->id_unit);
        return 0;
    }
    splx(s);
    
    printf("zwl%d: PCCARD card probed\n", id->id_unit);
    return WL_IO_REGION;
}

/*
 *  Initialize the device - called from Slot manager.
 *  Return ENXIO if success, or 0 if failed
 */
static int zwl_init_pccard(struct pccard_devinfo *devi)
{
    int unit = devi->isahd.id_unit;

    /* validate unit number. */
    if (unit >= NZWL) {
        printf("zwl%d: Only assign %d units when compiling\n", unit, NZWL);
        return(ENODEV);
    }
    
    /* io base from isa_device, but irq from pccard_devinfo.slot? ref:slot.h */
    printf("zwl%d: PCCARD base=0x%3.3x, irq=%d\n", 
            unit, devi->isahd.id_iobase, devi->slt->irq);
    
    devi->isahd.id_irq == devi->slt->irq;
    if (zwl_probe_pccard(&devi->isahd) == 0)
        return(ENXIO);

    if (zwl_attach(&devi->isahd) == 0)
        return(ENXIO);

    return(0);
}

/*
 *  zwl_unload_pccard - unload the driver and clear the table.
 *  XXX TODO:
 *  This is usually called when the card is ejected, but
 *  can be caused by a modunload of a controller driver.
 *  The idea is to reset the driver's view of the device
 *  and ensure that any driver entry points such as
 *  read and write do not hang.
 */
static void zwl_unload_pccard(struct pccard_devinfo *devi)
{
    struct zwl_softc *sc = &zwl_softc[devi->isahd.id_unit];
    struct ifnet *ifp = &sc->sc_arpcom.ac_if;

    if (sc->gone) {
        printf("zwl%d: already unloaded\n", devi->isahd.id_unit);
        return;
    }

    ifp->if_flags &= ~IFF_RUNNING;
/*    if_down(ifp); */
    sc->gone = 1;

    printf("zwl%d: unload\n", devi->isahd.id_unit);
}

/*
 *  zwl_intr_pccard - Shared interrupt called from
 *  front end of PC-Card handler.
 */
static int zwl_intr_pccard(struct pccard_devinfo *devi)
{
    zwlintr(devi->isahd.id_unit);
    return(1);
}

#else   /* NCARD = 0 */
#   warning "************************"
#   warning "* WL24xx PCMCIA is off *"
#   warning "************************"
#endif  /* NCARD > 0 */
